Index: binutils/gas/config/tc-i370.c
diff -u binutils/gas/config/tc-i370.c:1.51 binutils/gas/config/tc-i370.c:1.54
--- binutils/gas/config/tc-i370.c:1.51	Tue Jun  1 19:01:22 1999
+++ binutils/gas/config/tc-i370.c	Mon Jun 28 19:00:03 1999
@@ -105,7 +105,7 @@
 static void i370_ltorg PARAMS ((int));
 static void i370_using PARAMS ((int));
 static void i370_drop PARAMS ((int));
-static void i370_make_relative PARAMS ((expressionS *exp));
+static void i370_make_relative PARAMS ((expressionS *exp, expressionS *baseaddr));
 
 #ifdef OBJ_ELF
 static bfd_reloc_code_real_type i370_elf_suffix PARAMS ((char **, expressionS *));
@@ -385,11 +385,34 @@
    I370_OPCODE_32 or I370_OPCODE_64.  */
 static int i370_size = I370_OPCODE_32;
 
-/* The base register to use for opcode with optional operands */ 
-static int i370_using_regno = -1;
+/* The base register to use for opcode with optional operands.
+ * We define two of these: "text" and "other".  Normally, "text"
+ * would get used in the .text section for branches, while "other"
+ * gets used in the .data section for address constants.  
+ *
+ * The idea of a second base register in a different section
+ * is foreign to the usual HLASM-style semantics; however, it
+ * allows us to provide support for dynamically loaded libraries,
+ * by allowing us to place address constants in a section other 
+ * than the text section. The "other" section need not be the 
+ * .data section, it can be any section that isn't the .text section.
+ *
+ * Note that HLASM defines a multiple, concurrent .using semantic
+ * that we do not: in calculating offsets, it uses either the most 
+ * recent .using directive, or the one with the smallest displacement.
+ * This allows HLASM to support a quasi-block-scope-like behaviour.
+ * Handy for people writing assembly by hand ... but not supported 
+ * by us.
+ */ 
+static int i370_using_text_regno = -1;
+static int i370_using_other_regno = -1;
 
 /* The base address for address literals */
-static expressionS i370_using_baseaddr;
+static expressionS i370_using_text_baseaddr;
+static expressionS i370_using_other_baseaddr;
+
+/* the "other" section, used only for syntax error detection */
+static segT i370_other_section = undefined_section;
 
 /* Opcode hash table.  */
 static struct hash_control *i370_hash;
@@ -1030,23 +1053,23 @@
 i370_dsect (xxx)
      int xxx; 
 {
-  as_tsktsk ("dsect not supported");
-}
-
-static void
-i370_csect (xxx)
-     int xxx; 
-{
   char *save_line = input_line_pointer;
-  static char section[] = ".text\n";
+  static char section[] = ".data\n";
 
-  /* Just pretend this is .section .text */
+  /* Just pretend this is .section .data */
   input_line_pointer = section;
   obj_elf_section (xxx);
 
   input_line_pointer = save_line;
 }
 
+static void
+i370_csect (xxx)
+     int xxx; 
+{
+  as_tsktsk ("csect not supported");
+}
+
 
 /* DC Define Const  is only partially supported.
  * For samplecode on what to do, look at i370_elf_cons() above.
@@ -1452,7 +1475,14 @@
   else if (1 == sz) exx->X_add_symbol = byte_poolP;
   exx->X_add_number = offset_in_pool;
   exx->X_op_symbol = NULL;
-  i370_make_relative (exx);
+
+  /* If the user has set up a base reg in another section,
+   * use that; otherwise use the text section.  */
+  if (0 < i370_using_other_regno) {
+    i370_make_relative (exx, &i370_using_other_baseaddr);
+  } else {
+    i370_make_relative (exx, &i370_using_text_baseaddr);
+  }
 }
  
 /* The symbol setup for the literal pool is done in two steps.  First,
@@ -1566,7 +1596,12 @@
   expression (exx);
 
   /* OK, now we have to subtract the "using" location  */
-  i370_make_relative (exx);
+  /* normally branches appear in the text secion only... */
+  if (text_section == now_seg || 0 > i370_using_other_regno) {
+    i370_make_relative (exx, &i370_using_text_baseaddr);
+  } else {
+    i370_make_relative (exx, &i370_using_other_baseaddr);
+  }
 
   /* put the * back */
   if (dot) *dot = '*';
@@ -1728,6 +1763,16 @@
   int biggest_align = 0;
   char pool_name[20];
 
+  if (now_seg != text_section &&
+      i370_other_section == undefined_section) {
+    as_bad (".ltorg without prior .using in section %s",
+      now_seg->name);
+  }
+  if (now_seg != text_section &&
+      i370_other_section != now_seg ) {
+    as_bad (".ltorg in section %s paired to .using in section %s",
+      now_seg->name, i370_other_section->name);
+  }
   if ((NULL == longlong_poolP) &&
       (NULL == word_poolP) &&
       (NULL == short_poolP) &&
@@ -1773,6 +1818,7 @@
        default:
           as_bad ("bad literal size\n");
     }
+    if (NULL == current_poolP) continue;
     sprintf (pool_name, ".LITP%01d%06d", litsize, lit_pool_num);
     symbol_locate (current_poolP, pool_name, now_seg,
                    (valueT) frag_now_fix (), frag_now);
@@ -1825,7 +1871,7 @@
 i370_using (ignore)
      int ignore;
 {
-  expressionS ex;
+  expressionS ex, baseaddr;
   int iregno;
   char *star;
 
@@ -1836,11 +1882,11 @@
 
   /* the first arg to using will usually be ".", but it can
    * be a more complex exprsssion too ... */
-  expression (&i370_using_baseaddr);
+  expression (&baseaddr);
   if (star) *star = '*';
-  if ((O_constant != i370_using_baseaddr.X_op) &&
-      (O_symbol   != i370_using_baseaddr.X_op) &&
-      (O_uminus   != i370_using_baseaddr.X_op))
+  if ((O_constant != baseaddr.X_op) &&
+      (O_symbol   != baseaddr.X_op) &&
+      (O_uminus   != baseaddr.X_op))
   {
     as_bad (".using: base address expression illegal or too complex");
   }
@@ -1852,19 +1898,14 @@
   demand_empty_rest_of_line ();
   iregno = ex.X_add_number;
 
-  /* if we were strict, we could check that the current 
-   * using was -1 and if not, flag it as a warning ... */
-  if (-1 != i370_using_regno) {
-    // In fact, there can be multiple usings outstanding;
-    // the most recent or the one with the least offset is
-    // the one that gets used.  We don't support this ...
-    // but don't print the error mesg,  the common
-    // usage doesn't mess things up but does print this msg.
-    //     as_tsktsk (".using %d without dropping %d first", 
-    //       iregno, i370_using_regno);
+  if (text_section == now_seg) {
+    i370_using_text_baseaddr = baseaddr;
+    i370_using_text_regno = iregno;
+  } else {
+    i370_using_other_baseaddr = baseaddr;
+    i370_using_other_regno = iregno;
+    i370_other_section = now_seg;
   }
-
-  i370_using_regno = iregno;
 }
 
 static void
@@ -1878,33 +1919,51 @@
   demand_empty_rest_of_line ();
   iregno = ex.X_add_number;
 
-  if (iregno != i370_using_regno) {
-    as_bad ("droping register %d does not match using register %d", 
-      iregno, i370_using_regno);
+  if (text_section == now_seg) {
+    if (iregno != i370_using_text_regno) {
+      as_bad ("droping register %d in section %s does not match using register %d", 
+        iregno, now_seg->name, i370_using_text_regno);
+    }
+    i370_using_text_regno = -1;
+    i370_using_text_baseaddr.X_op = O_absent;
+  } else {
+    if (iregno != i370_using_other_regno) {
+      as_bad ("droping register %d in section %s does not match using register %d", 
+        iregno, now_seg->name, i370_using_other_regno);
+    }
+    if (i370_other_section != now_seg) {
+      as_bad ("droping register %d in section %s previously used in section %s",
+        iregno, now_seg->name, i370_other_section->name);
+    }
+    i370_using_other_regno = -1;
+    i370_using_other_baseaddr.X_op = O_absent;
+    i370_other_section = undefined_section;
   }
-  i370_using_regno = -1;
-  i370_using_baseaddr.X_op = O_absent;
 }
 
+/* Make the first argument an address-relative expression
+ * by subtracting the second argument.
+ */
 static void
-i370_make_relative (expressionS *exx)
+i370_make_relative (expressionS *exx, expressionS *baseaddr)
 {
-  if (O_constant == i370_using_baseaddr.X_op) 
+
+  if (O_constant == baseaddr->X_op) 
   {
      exx->X_op = O_symbol;
-     exx->X_add_number -= i370_using_baseaddr.X_add_number;
+     exx->X_add_number -= baseaddr->X_add_number;
   } else
-  if (O_symbol == i370_using_baseaddr.X_op) 
+  if (O_symbol == baseaddr->X_op) 
   {
      exx->X_op = O_subtract;
-     exx->X_op_symbol = i370_using_baseaddr.X_add_symbol;
-     exx->X_add_number -= i370_using_baseaddr.X_add_number;
+     exx->X_op_symbol = baseaddr->X_add_symbol;
+     exx->X_add_number -= baseaddr->X_add_number;
   } else
-  if (O_uminus == i370_using_baseaddr.X_op) 
+  if (O_uminus == baseaddr->X_op) 
   {
      exx->X_op = O_add;
-     exx->X_op_symbol = i370_using_baseaddr.X_add_symbol;
-     exx->X_add_number += i370_using_baseaddr.X_add_number;
+     exx->X_op_symbol = baseaddr->X_add_symbol;
+     exx->X_add_number += baseaddr->X_add_number;
    } else 
    {
       as_bad ("Missing or bad .using directive");
@@ -1936,6 +1995,7 @@
   const unsigned char *opindex_ptr;
   int have_optional_index, have_optional_basereg, have_optional_reg;
   int skip_optional_index, skip_optional_basereg, skip_optional_reg;
+  int use_text=0, use_other=0;
   int off_by_one;
   int need_paren = 0;
   struct i370_fixup fixups[MAX_INSN_FIXUPS];
@@ -2083,10 +2143,25 @@
       if (skip_optional_basereg && 
           ((operand->flags & I370_OPERAND_BASE) != 0))
         {
-          if (0 > i370_using_regno) {
+          int basereg = -1;
+          if (use_text) {
+            if (text_section == now_seg || 0 > i370_using_other_regno) {
+              basereg = i370_using_text_regno;
+            } else {
+              basereg = i370_using_other_regno;
+            } 
+          }
+          else if (use_other) {
+            if (0 > i370_using_other_regno) {
+              basereg = i370_using_text_regno;
+            } else {
+              basereg = i370_using_other_regno;
+            } 
+          }
+          if (0 > basereg) {
             as_bad ("not using any base register");
           }
-          insn = i370_insert_operand (insn, operand, i370_using_regno,
+          insn = i370_insert_operand (insn, operand, basereg,
         			     (char *) NULL, 0);
           continue;
         }
@@ -2122,8 +2197,10 @@
         }
 
       /* check for a address constant expression */
-      else if (i370_addr_cons (&ex)) {}
-      else if (i370_addr_offset (&ex)) {}
+      /* We will put PSW-relative addresses in the text section,
+       * and adress literals in the .data (or other) section. */
+      else if (i370_addr_cons (&ex)) { use_other=1; }
+      else if (i370_addr_offset (&ex)) { use_text=1; }
       else expression (&ex);
 
       str = input_line_pointer;
Index: binutils/gas/doc/c-i370.texi
diff -u /dev/null binutils/gas/doc/c-i370.texi:1.2
--- /dev/null	Tue Jul  6 18:52:30 1999
+++ binutils/gas/doc/c-i370.texi	Mon Jun 28 19:53:41 1999
@@ -0,0 +1,200 @@
+@c Copyright (C) 1999 Free Software Foundation, Inc.
+@c This is part of the GAS manual.
+@c For copying conditions, see the file as.texinfo.
+@ifset GENERIC
+@page
+@node ESA/390-Dependent
+@chapter ESA/390 Dependent Features
+@end ifset
+@ifclear GENERIC
+@node Machine Dependencies
+@chapter ESA/390 Dependent Features
+@end ifclear
+
+@cindex i370 support
+@cindex ESA/390 support
+
+@menu
+* ESA/390 Notes::                Notes
+* ESA/390 Options::              Options
+* ESA/390 Syntax::               Syntax
+* ESA/390 Floating Point::       Floating Point
+* ESA/390 Directives::           ESA/390 Machine Directives
+* ESA/390 Opcodes::              Opcodes
+@end menu
+
+@node ESA/390 Notes
+@section Notes
+The ESA/390 @code{@value{AS}} port is currently intended to be a back-end
+for the @sc{gnu} @sc{cc} compiler.  It is not HLASM compatible, although
+it does support a subset of some of the HLASM directives.  The only 
+supported binary file format is ELF; none of the usual MVS/VM/OE/USS 
+object file formats, such as ESD or XSD, are supported.
+
+When used with the @sc{gnu} @sc{cc} compiler, the ESA/390 @code{@value{AS}}
+will produce correct, fully relocated, functional binaries, and has been 
+used to compile and execute large projects.  However, many aspects should 
+still be considered experimental; these include shared library support, 
+dynamically loadable objects, and any relocation other than the 31-bit 
+relocation.
+
+@node ESA/390 Options
+@section Options
+@code{@value{AS}} has no machine-dependent command-line options for the ESA/390.
+
+@cindex ESA/390 Syntax
+@node ESA/390 Syntax
+@section Syntax
+The opcode/operand syntax follows the ESA/390 Principles of Operation
+manual; assembler directives and general syntax are loosely based on the 
+prevailing AT&T/SVR4/ELF/Solaris style notation.  HLASM-style directives
+are @emph{not} supported for the most part, with the exception of those 
+described herein.
+
+A leading dot in front of directives is optional, and the case of
+directives is ignored; thus for example, .using and USING have the same
+effect.
+
+A colon may immediately follow a label definition.  This is
+simply for compatibility with how most assembly language programmers
+write code.
+
+@samp{#} is the line comment character.
+
+@samp{;} can be used instead of a newline to separate statements.
+
+Since @samp{$} has no special meaning, you may use it in symbol names.
+
+Registers can be given the symbolic names r0..r15, fp0, fp2, fp4, fp6.
+By using thesse symbolic names, @code{@value{AS}} can detect simple 
+syntax errors. The name rarg or r.arg is a synonym for r11, rtca or r.tca
+for r12, sp, r.sp, dsa r.dsa for r13, lr or r.lr for r14, rbase or r.base 
+for r3 and rpgt or r.pgt for r4.
+
+@samp{*} is the current location counter.  Unlike @samp{.} it is always
+relative to the last USING directive.  Note that this means that 
+expressions cannot use multiplication, as any occurence of @samp{*}
+will be interpreted as a location counter.
+
+All labels are relative to the last USING.  Thus, branches to a label 
+always imply the use of base+displacement.
+
+Many of the usual forms of address constants / address literals 
+are supported.  Thus,
+@example
+	.using	*,r3
+	L	r15,=A(some_routine)
+	LM	r6,r7,=V(some_longlong_extern)
+	A	r1,=F'12'
+	AH	r0,=H'42'
+	ME	r6,=E'3.1416'
+	MD	r6,=D'3.14159265358979'
+	O	r6,=XL4'cacad0d0'
+	.ltorg
+@end example
+should all behave as expected: that is, an entry in the literal
+pool will be created (or reused if it already exists), and the 
+instruction operands will be the displacement into the literal pool
+using the current base register (as last declared with the @code{.using}
+directive).
+
+@node ESA/390 Floating Point
+@section Floating Point
+@cindex floating point, ESA/390 (@sc{ieee})
+@cindex ESA/390 floating point (@sc{ieee})
+The assembler generates only @sc{ieee} floating-point numbers.  The older
+floiating point formats are not supported.
+
+
+@node ESA/390 Directives
+@section ESA/390 Assembler Directives
+
+@code{@value{AS}} for the ESA/390 supports all of the standard ELF/SVR4 
+assembler directives that are documented in the main part of this
+documentation.  Several additional directives are supported in order
+to implement the ESA/390 addressing model.  The most important of these
+are @code{.using} and @code{.ltorg}
+
+@cindex ESA/390-only directives
+These are the additional directives in @code{@value{AS}} for the ESA/390:
+
+@table @code
+@item .dc 
+A small subset of the usual DC directive is supported.
+
+@item .drop @var{regno}
+Stop using @var{regno} as the base register.  The @var{regno} must
+have been previously declared with a @code{.using} directive in the
+same section as the current section.
+
+@item .ebcdic @var{string}
+Emit the EBCDIC equivalent of the indicated string.  The emitted string
+will be null terminated.  Note that the directives @code{.string} etc. emit
+ascii strings by default.
+
+@item EQU 
+The standard HLASM-style EQU directive is not supported; however, the 
+standard @code{@value{AS}} directive .equ can be used to the same effect.
+
+@item .ltorg 
+Dump the literal pool accumulated so far; begin a new literal pool.
+The literal pool will be written in the current section; in order to
+generate correct assembly, a @code{.using} must have been previously
+specified in the same section.
+
+@item .using @var{expr},@var{regno}
+Use @var{regno} as the base register for all subsequent RX, RS, and SS form
+instructions. The @var{expr} will be evaluated to obtain the base address;
+usually, @var{expr} will merely be @samp{*}.
+
+This assembler allows two @code{.using} directives to be simultaneously
+outstanding, one in the @code{.text} section, and one in another section 
+(typically, the @code{.data} section).  This feature allows 
+dynamically loaded objects to be implemented in a relatively 
+straightforward way.  A @code{.using} directive must always be specified 
+in the @code{.text} section; this will specify the base register that
+will be used for branches in the @code{.text} section.  A second
+@code{.using} may be specified in another section; this will specify
+the base register that is used for non-label address literals.
+When a second @code{.using} is specified, then the subsequent
+@code{.ltorg} must be put in the same section; otherwise an error will 
+result.
+
+Thus, for example, the following code uses @code{r3} to address branch 
+targets and @code{r4} to address the literal pool, which has been written 
+to the @code{.data} section.  The is, the constants @code{=A(some_routine)},
+@code{=H'42'} and @code{=E'3.1416'} will all appear in the @code{.data}
+section.
+
+@example
+.data
+	.using  LITPOOL,r4
+.text
+	BASR	r3,0
+	.using	*,r3
+        B       START
+	.long	LITPOOL
+START:
+	L	r4,4(,r3)
+	L	r15,=A(some_routine)
+	LTR	r15,r15
+	BNE	LABEL
+	AH	r0,=H'42'
+LABEL:
+	ME	r6,=E'3.1416'
+.data
+LITPOOL:
+	.ltorg
+@end example
+
+
+Note that this dual-@code{.using} directive semantics extends 
+and is not compatible with HLASM semantics.  Note that this assembler 
+directive does not support the full range of HLASM semantics.
+
+@end table
+
+@node ESA/390 Opcodes
+@section Opcodes
+For detailed information on the ESA/390 machine instruction set, see
+@cite{ESA/390 Principles of Operation} (IBM Publication Number DZ9AR004).
